/* S1 microprogram simulation

   Prabhas Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

   17 July 98
   s2 micro mpgm sim		13 Feb 2005
*/

#include "s2.h"

#define 	PRIVATE		static

char ms[MLEN][MWIDTH]; 		// micro store
int  mpc, here;				// micro PC

int  mbus,tbus,p1,p2,aout;  // internal bus, alu ports
int  rin,ry,ax,pcin,rdest;	// wires
int  clock;

sym mads[100];						// symbol for microads
int mwidth, mawidth, mcwidth, mlen;	// mpgm parameters
PRIVATE int udop[32], udxop[32];	// micro rom

/* test current ms is set at pos */
PRIVATE int set(int pos ){
  return( ms[here][pos] == '1' );
}
// lookup name-val pair in mads, 0 not found
PRIVATE int lookup(char *s){
	int i;
	i = 0;
	while(mads[i].name[0] != 0){
		if( eqs(mads[i].name,s) ) return mads[i].val;
		i++;
	}
	printf("not found %s in mads\n",s);
	return 0;
}

// put micro-address into rom : udop[], udxop[]
void initrom(void){
	udop[LDA] = lookup("lda");
	udop[LDD] = lookup("ldd");
	udop[LDI] = lookup("ldi");
	udop[STA] = lookup("sta");
	udop[STD] = lookup("std");
	udop[ADDI] = lookup("opim");
	udop[SUBI] = lookup("opim");
	udop[MULI] = lookup("opim");
	udop[DIVI] = lookup("opim");
	udop[ANDI] = lookup("opim");
	udop[ORI] = lookup("opim");
	udop[XORI] = lookup("opim");
	udop[JAL] = lookup("jal");
	udop[JMP] = lookup("jcc");
	udop[XL] = lookup("xl");
	udop[XD] = lookup("xd");
	udxop[ADD] = lookup("add");
	udxop[SUB] = lookup("bop");
	udxop[MUL] = lookup("bop");
	udxop[DIV] = lookup("bop");
	udxop[AND] = lookup("bop");
	udxop[OR] = lookup("bop");
	udxop[XOR] = lookup("bop");
	udxop[SHL] = lookup("bop");
	udxop[SHR] = lookup("bop");
	udxop[LDX] = lookup("ldx");
	udxop[STX] = lookup("stx");
	udxop[JR] = lookup("jr");
	udxop[TRAP] = lookup("trap");
	udxop[XX] = lookup("xx");
}

/* decode instruction to microprogram address */

PRIVATE int udecode(void){
  int a;
  a = 0;
  if( IRop() != XOP ) a = udop[IRop()];
  else if( IRxop() <= XX ) a = udxop[IRxop()];
  if( a == 0 )
  	error("undefined opcode");
  return a;
}

PRIVATE int ugoto(void){	/* decode Goto field */
  int i, a = 0, w = 1;
  for(i=mwidth-1; i>=mcwidth; i--){
	if( set(i) )   a += w;
	w += w;
  }
  return a;
}

// run events of each field
// execute one inst. (a number of microsteps)
void run2(void){
  savePc = Pc;
  while(1) {
	here = mpc;
    clock++;
/*
	if(clock > 100) {
		runflag = 0;
		break;
	}
*/
    if( mpc == a_trap ) trap(IRr1(),IRr2());

	if( set(s_x_r1) ) 	p1 = R[IRr1()];
    if( set(s_x_r2) ) 	p1 = R[IRr2()];
	if( set(s_x_r3) ) 	p1 = R[IRr3()];
    if( set(s_y_r1) ) 	ry = R[IRr1()];
	if( set(s_y_r2) ) 	ry = R[IRr2()];
	if( set(s_y_r3) ) 	ry = R[IRr3()];
    if( set(s_z_ry) )   p2 = ry;
    if( set(s_z_im) ) 	p2 = IRdisp();
	if( set(s_z_0) ) 	p2 = 0;
	if( set(s_z_1) ) 	p2 = 1;
	if( set(s_z_2) ) 	p2 = 2;
	if( set(s_z_4) ) 	p2 = 4;

	if( set(s_mr) ) 	mbus = M[Mar];

	if( set(s_r_mbus) ) rin = mbus;
    if( set(s_r_pc) ) 	rin = Pc;
    if( set(s_r_tbus) ) rin = tbus;
	if( set(s_r_ads) )  rin = IRads();
    if( set(s_a_ads) ) 	ax = IRads();
    if( set(s_a_pc) ) 	ax = Pc;
    if( set(s_a_tbus) ) ax = tbus;
    if( set(s_j_pc1) ) 	pcin = Pc+1;
    if( set(s_j_ads) ) 	pcin = IRads();
    if( set(s_j_p1) ) 	pcin = p1;

    if( set(s_alu_add) )  aout = alu(FADD,p1,p2);
	if( set(s_alu_sub))   aout = alu(FSUB,p1,p2);
    if( set(s_alu_mul) )  aout = alu(FMUL,p1,p2);
	if( set(s_alu_div))   aout = alu(FDIV,p1,p2);
    if( set(s_alu_and) )  aout = alu(FAND,p1,p2);
	if( set(s_alu_or))    aout = alu(FOR,p1,p2);
    if( set(s_alu_xor) )  aout = alu(FXOR,p1,p2);
	if( set(s_alu_shl))   aout = alu(FSHL,p1,p2);
    if( set(s_alu_shr) )  aout = alu(FSHR,p1,p2);
	if( set(s_alu_pass1)) aout = alu(FP1,p1,p2);
	if( set(s_alu_pass2)) aout = alu(FP2,p1,p2);
	if( set(s_alu_op) )   aout = alu(mapop(IRop(),IRxop()),p1,p2);

    if( set(s_mw) ) 	M[Mar] = tbus;
	if( set(s_d_r1) )	rdest = IRr1();
	if( set(s_d_r2) )	rdest = IRr2();
	if( set(s_d_r3) )	rdest = IRr3();

    if( set(s_lpc) ) 	Pc = pcin;
	if( set(s_lir) ) 	Ir = mbus;
    if( set(s_lmar) ) 	Mar = ax;
	if( set(s_lr) )		wR(rdest,rin);
    if( set(s_lt) ) 	T = aout;

	tbus = T;

	mpc++;
	if( set(s_jmp))		mpc = ugoto();
	if( set(s_jF)) 		if( !testCc(IRr1()) ) mpc = ugoto();
	if( set(s_jT)) 		if( testCc(IRr1()) ) mpc = ugoto();
	if( set(s_jz)) 		if( Z == 1 ) mpc = ugoto();
	if( set(s_jnz))		if( Z == 0 ) mpc = ugoto();
	if( set(s_decode))	mpc = udecode();

	if(disp) show();
	if(mpc == 0) break;
  }
}

